from abc import abstractmethod
import numpy as np
from scipy.integrate import odeint
import warnings

from gym import core
from gym.utils import seeding
from gym import spaces


class WrongObservationError(Exception):
    def __init__(self, msg, observation, observation_space):
        msg_ext = self.get_wrong_observation(observation, observation_space)
        super().__init__(msg + msg_ext)

    def get_wrong_observation(
        self, observation: dict, observation_space: spaces.Dict
    ) -> str:
        msg_ext = ": "
        for key in observation.keys():
            if not observation_space[key].contains(observation[key]):
                msg_ext += "Error in " + key
                for i, val in enumerate(observation[key]):
                    if val < observation_space[key].low[i]:
                        msg_ext += (
                            f"[{i}]: {val} < {observation_space[key].low[i]}"
                        )
                    elif val > observation_space[key].high[i]:
                        msg_ext += (
                            f"[{i}]: {val} > {observation_space[key].high[i]}"
                        )
        return msg_ext


class PlanarEnv(core.Env):
    def __init__(self, render: bool = False, dt=0.01):
        self._viewer = None
        self._state = {'x': None, 'xdot': None}
        self._sensor_state = None
        self.seed()
        self._dt = dt
        self._t = 0.0
        self._render = render
        self._obsts = []
        self._goals = []
        self._sensors = []
        self.observation_space = None
        self._n = None

    @property
    def n(self):
        return self._n

    @n.setter
    def n(self, n):
        self._n = n

    @n.deleter
    def n(self):
        del self._n

    @abstractmethod
    def set_spaces(self):
        pass

    def dt(self):
        return self._dt

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def add_obstacle(self, obst):
        self._obsts.append(obst)

    def add_goal(self, goal):
        self._goals.append(goal)

    def add_sensor(self, sensor):
        self._sensors.append(sensor)
        observation_space_dict = dict(self.observation_space.spaces)
        print(sensor.name)
        observation_space_dict[sensor.name] = sensor.observation_space()
        self.observation_space = spaces.Dict(observation_space_dict)

    def t(self):
        return self._t

    def reset_common(self):
        self._obsts = []
        self._goals = []
        self._sensors = []
        self._t = 0.0

    def reset(self, pos: np.ndarray = None, vel: np.ndarray = None) -> dict:
        self.reset_common()
        if not isinstance(pos, np.ndarray) or not pos.size == self._n:
            pos = np.zeros(self._n)
        if not isinstance(vel, np.ndarray) or not vel.size == self._n:
            vel = np.zeros(self._n)
        self._state = {"x": pos, "xdot": vel}
        self._sensor_state = {}
        return self._get_ob()

    def step(self, action: np.ndarray) -> tuple:
        self._action = action
        self.integrate()
        for sensor in self._sensors:
            self._sensor_state[sensor.name] = sensor.sense(
                self._state, self._goals, self._obsts, self.t()
            )
        terminal = self._terminal()
        reward = self._reward()
        if self._render:
            self.render()
        return (self._get_ob(), reward, terminal, {})

    @abstractmethod
    def _reward(self):
        pass

    def _get_ob(self):
        observation = dict(self._state)
        observation.update(self._sensor_state)
        if not self.observation_space.contains(observation):
            err = WrongObservationError(
                "The observation does not fit the defined observation space",
                observation,
                self.observation_space,
            )
            warnings.warn(str(err))
        return observation

    @abstractmethod
    def _terminal(self):
        pass

    @abstractmethod
    def continuous_dynamics(self, x, t):
        pass

    def integrate(self):
        self._t += self.dt()
        t = np.arange(0, 2 * self._dt, self._dt)
        x0 = np.concatenate((self._state["x"], self._state["xdot"]))
        ynext = odeint(self.continuous_dynamics, x0, t)
        self._state["x"] = ynext[1][0 : self._n]
        self._state["xdot"] = ynext[1][self._n : 2 * self._n]

    @abstractmethod
    def render(self, mode="human"):
        pass

    def render_common(self, bounds):
        from gym.envs.classic_control import (
            rendering,
        )  # pylint: disable=import-outside-toplevel

        if self._state is None:
            return None
        if self._viewer is None:
            if isinstance(bounds, list):
                self._viewer = rendering.Viewer(500, 500)
                self._viewer.set_bounds(
                    -bounds[0], bounds[1], -bounds[1], bounds[1]
                )
            elif isinstance(bounds, dict):
                ratio = (bounds["pos"]["high"][0] - bounds["pos"]["low"][0]) / (
                    bounds["pos"]["high"][1] - bounds["pos"]["low"][1]
                )
                if ratio > 1:
                    window_size = (1000, int(1000 / ratio))
                else:
                    window_size = (int(ratio * 1000), 1000)
                self._viewer = rendering.Viewer(window_size[0], window_size[1])
                self._viewer.set_bounds(
                    bounds["pos"]["low"][0],
                    bounds["pos"]["high"][0],
                    bounds["pos"]["low"][1],
                    bounds["pos"]["high"][1],
                )
        for obst in self._obsts:
            obst.renderGym(self._viewer, rendering, t=self.t())
        for goal in self._goals:
            goal.renderGym(self._viewer, rendering, t=self.t())

    def close(self):
        if self._viewer:
            self._viewer.close()
            self._viewer = None



# pylint: disable-all
import casadi as ca
from casadi import cos, sin
import numpy as np
from sys import argv


# create variables
def main():
    # n = int(argv[1])
    n = 15
    print(
        "Creating dynamics for pendulum with "
        + str(n)
        + " joints, point masses"
    )
    name = "./" + str(n) + "_Reacher.casadi"
    dynamics, _, _ = createDynamics(n)
    dynamics.save(name)


def createDynamics(n):
    q = ca.SX.sym("q", n)
    qdot = ca.SX.sym("qdot", n)
    qddot = ca.SX.sym("qddot", n)
    tau = ca.SX.sym("tau", n)
    l = ca.SX.sym("l", n)
    m = ca.SX.sym("m", n)
    k = ca.SX.sym("k", 1)
    g = ca.SX.sym("g", 1)
    fk = ca.SX.sym("x", (n, 2))
    v = ca.SX.sym("v", (n, 2))
    v_2 = ca.SX.sym("v_2", n)

    # velocities and positions
    v[0, 0] = -1 * sin(q[0]) * l[0] * qdot[0]
    v[0, 1] = cos(q[0]) * l[0] * qdot[0]
    fk[0, 0] = cos(q[0]) * l[0]
    fk[0, 1] = sin(q[0]) * l[0]

    for i in range(1, n):
        angle = 0.0
        for j in range(i + 1):
            angle += q[j]
        v[i, 0] = v[i - 1, 0] - sin(angle) * l[i] * qdot[i]
        v[i, 1] = v[i - 1, 1] + cos(angle) * l[i] * qdot[i]
        fk[i, 0] = fk[i - 1, 0] + cos(angle) * l[i]
        fk[i, 1] = fk[i - 1, 1] + sin(angle) * l[i]

    for i in range(n):
        v_2[i] = ca.norm_2(v[i, :]) ** 2

    # kinetic energy
    T = sum([0.5 * m[i] * v_2[i] for i in range(n)])

    # potential energy
    V = sum([g * m[i] * fk[i, 1] for i in range(n)])

    # lagrangian
    L = T - V

    # derivatives
    dL_dq = ca.gradient(L, q)
    dL_dqdot = ca.gradient(L, qdot)

    d2L_dq2 = ca.jacobian(dL_dq, q)
    d2L_dqdqdot = ca.jacobian(dL_dq, qdot)
    d2L_dqdot2 = ca.jacobian(dL_dqdot, qdot)

    M = d2L_dqdot2
    F = d2L_dqdqdot
    f = dL_dq
    # damping
    K = np.identity(n) * k

    tau_forward = (
        ca.mtimes(M, qddot) + ca.mtimes(F, qdot) + ca.mtimes(K, qdot) - f
    )
    # equation of motion
    # M * q_ddot + F q_dot - f = tau
    # augmented for first order system
    tau_aug = ca.vertcat(np.zeros((n, 1)), tau)
    f_aug = ca.vertcat(np.zeros((n, 1)), f)
    F_aug = ca.horzcat(np.zeros((2 * n, n)), ca.vertcat(-1 * np.identity(n), F))
    K_aug = ca.horzcat(np.zeros((2 * n, n)), ca.vertcat(0 * np.identity(n), K))
    M_aug = ca.vertcat(
        ca.horzcat(np.identity(n), np.zeros((n, n))),
        ca.horzcat(np.zeros((n, n)), M),
    )

    x = ca.vertcat(q, qdot)

    rhs = tau_aug + f_aug - ca.mtimes(F_aug, x) - ca.mtimes(K_aug, x)
    xdot = ca.solve(M_aug, rhs)

    dynamics = ca.Function("dynamics", [q, qdot, l, m, g, k, tau], [xdot])
    fk = ca.Function("fk", [q, l], [fk])
    tau_fun = ca.Function("tau", [q, qdot, l, m, g, k, qddot], [tau_forward])
    return dynamics, fk, tau_fun


if __name__ == "__main__":
    main()



import numpy as np
from gym import spaces

from planarenvs.n_link_reacher.envs.n_link_reacher_env import NLinkReacherEnv
from planarenvs.n_link_reacher.resources.createDynamics import createDynamics


class NLinkTorReacherEnv(NLinkReacherEnv):
    def __init__(self, render=False, n=2, dt=0.01, friction=0.0):
        super().__init__(render, n, dt)
        self._dynamics_fun, _, _ = createDynamics(self._n)
        self._friction = friction
        self._g = 10

    def set_spaces(self):
        self.observation_space = spaces.Dict(
            {
                "x": spaces.Box(
                    low=-self._lim_up_pos,
                    high=self._lim_up_pos,
                    dtype=np.float64
                ),
                "xdot": spaces.Box(
                    low=-self._lim_up_vel,
                    high=self._lim_up_vel,
                    dtype=np.float64
                ),
            }
        )
        self.action_space = spaces.Box(
            low=-self._lim_up_tor, high=self._lim_up_tor, dtype=np.float64
        )

    def continuous_dynamics(self, x, t):
        q = x[0 : self._n]
        qdot = x[self._n : self._n * 2]
        l = [self.LINK_LENGTH for i in range(self._n)]
        m = [self.LINK_MASS for i in range(self._n)]
        tau = self._action
        acc = np.array(
            self._dynamics_fun(q, qdot, l, m, self._g, self._friction, tau)
        )[:, 0]
        return acc



import numpy as np
from gym import spaces


class VelEnv(object):
    def set_spaces(self):
        self.observation_space = spaces.Dict(
            {
                "x": spaces.Box(
                    low=-self._lim_up_pos, high=self._lim_up_pos, dtype=np.float64
                ),
                "xdot": spaces.Box(
                    low=-self._lim_up_vel, high=self._lim_up_vel, dtype=np.float64
                ),
            }
        )
        self.action_space = spaces.Box(
            low=-self._lim_up_vel, high=self._lim_up_vel, dtype=np.float64
        )

    def integrate(self):
        super().integrate()
        self._state["xdot"] = self._action

    def continuous_dynamics(self, x, t):  # pylint: disable=unused-argument
        vel = self._action
        acc = np.zeros(self._n)
        xdot = np.concatenate((vel, acc))
        return xdot
